/*-
 * #%L
 * Smile CDR - CDR
 * %%
 * Copyright (C) 2016 - 2024 Smile CDR, Inc.
 * %%
 * All rights reserved.
 * #L%
 */
package com.smilecdr.demo.fhirgateway;

import ca.cdr.api.fhir.interceptor.CdrHook;
import ca.cdr.api.fhir.interceptor.CdrPointcut;
import ca.cdr.api.fhirgw.json.GatewayTargetJson;
import ca.cdr.api.fhirgw.model.ISearchResultsAccumulator;
import ca.cdr.api.fhirgw.model.SearchRequest;
import ca.cdr.api.fhirgw.model.SearchResponse;
import ca.uhn.fhir.interceptor.api.Interceptor;
import ca.uhn.fhir.rest.server.exceptions.InvalidRequestException;
import ca.uhn.fhir.rest.server.servlet.ServletRequestDetails;
import org.hl7.fhir.instance.model.api.IBaseResource;
import org.hl7.fhir.r4.model.IdType;
import org.hl7.fhir.r4.model.Observation;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.apache.commons.lang3.StringUtils.isBlank;

/**
 * This interceptor demonstrates a gateway search where the target uses different
 * resource IDs than the one(s) the client understands.
 */
@Interceptor
public class GatewaySearchUsingAlternateTargetIds {

	private final Map<String, String> myClientIdToTargetId;

	/**
	 * Constructor
	 */
	public GatewaySearchUsingAlternateTargetIds() {

		// For this example, we'll just use a hardcoded map to translate the
		// IDs we receive from the client into the IDs we'll use to communicate
		// with the target. Obviously in a real scenario this would probably
		// be a lookup to something external.
		myClientIdToTargetId = new HashMap<>();
		myClientIdToTargetId.put("100", "AAA");
		myClientIdToTargetId.put("200", "BBB");

	}

	/**
	 * This hook is called before the search is executed against the target server
	 */
	@CdrHook(CdrPointcut.FHIRGW_SEARCH_TARGET_PREINVOKE)
	public void preInvoke(SearchRequest theRequest, GatewayTargetJson theTarget, ISearchResultsAccumulator theISearchResultsAccumulator, ServletRequestDetails theRequestDetails) {

		// We assume there are two targets, one called "target1" and one called "target2". This
		// logic only applies to the second target.
		if (!theTarget.getId().equals("target2")) {
			return;
		}

		// This logic only applies to the Observation resource type
		if (!theRequest.getResourceType().equals("Observation")) {
			return;
		}

		// Look for the subject ID in either the "patient" or "subject"
		// URL parameters.
		String patientId = theRequest.getParameter("patient");
		if (isBlank(patientId)) {
			patientId = theRequest.getParameter("subject");
		}
		if (isBlank(patientId)) {
			throw new InvalidRequestException("No patient ID supplied in request");
		}

		// The ID supplied by the client can be in forms like "Patient/100" or just "100"
		// so we'll use a HAPI FHIR IdType structure to extract just the ID Part (100).
		patientId = new IdType(patientId).getIdPart();

		// Lookup the mapping for the patient ID in our map
		String newPatientId = myClientIdToTargetId.get(patientId);
		if (newPatientId == null) {
			throw new InvalidRequestException("No mapping found for patient " + patientId);
		}

		// Replace the existing search parameter with the mapped version
		theRequest.removeParameters("patient");
		theRequest.removeParameters("subject");
		theRequest.addParameter("patient", "Patient/" + newPatientId);

		// Store the original patient ID an an attribute so we can look it up later. Note that
		// we use the ID of the target in the attribute name just to avoid any conflicts if
		// multiple interceptors like this one are registered
		theRequestDetails.setAttribute("target2-observation-patientId", patientId);

	}

	/**
	 * This hook is called after the target server has responded, and can be used to manipulate
	 * the results
	 */
	@CdrHook(CdrPointcut.FHIRGW_SEARCH_TARGET_POSTINVOKE)
	public void postInvoke(GatewayTargetJson theTarget, SearchResponse theSearchResponse, ServletRequestDetails theRequestDetails) {

		// We assume there are two targets, one called "target1" and one called "target2". This
		// logic only applies to the second target.
		if (!theTarget.getId().equals("target2")) {
			return;
		}

		String patientId = (String) theRequestDetails.getAttribute("target2-observation-patientId");
		if (isBlank(patientId)) {
			return;
		}

		// Replace the patient ID in the results we received from "target2" with the
		// original ID the user had searched for. This is important because the
		// client won't be expecting to see references using the internal ID, and also
		// because the authorization service will only allow the client to access
		// patients in compartments it has access to.
		List<IBaseResource> results = theSearchResponse.getSearchResults();
		for (IBaseResource nextResult : results) {
			if (nextResult instanceof Observation) {
				Observation nextObservation = (Observation) nextResult;
				nextObservation.getSubject().setReference("Patient/" + patientId);
			}
		}

	}


}
