/*-
 * #%L
 * Smile CDR - CDR
 * %%
 * Copyright (C) 2016 - 2024 Smile CDR, Inc.
 * %%
 * All rights reserved.
 * #L%
 */
package com.smilecdr.demo.fhirstorage;

import ca.cdr.api.fhir.interceptor.CdrHook;
import ca.cdr.api.fhir.interceptor.CdrPointcut;
import ca.cdr.api.persistence.megascale.MegaScaleCredentialRequestJson;
import ca.cdr.api.persistence.megascale.MegaScaleCredentialResponseJson;
import ca.uhn.fhir.interceptor.api.Interceptor;
import ca.uhn.fhir.rest.server.exceptions.InvalidRequestException;
import org.apache.commons.lang3.StringUtils;

/**
 * This interceptor is invoked when initiating a request for a MegaScale
 * connection for a specific partition. It returns the associated JDBC
 * URL, username and password to use for that partition.
 *
 * Responses are cached, so this pointcut will not be repeatedly invoked
 * every time a partition is accessed and this interceptor is therefore
 * not considered performance critical if latency is incurred.
 */
@Interceptor
public class MegaScaleConnectionProvidingInterceptor {

	@CdrHook(CdrPointcut.STORAGE_MEGASCALE_PROVIDE_DB_INFO)
	public MegaScaleCredentialResponseJson provideMegaScaleCredentials(MegaScaleCredentialRequestJson theRequest) {

		MegaScaleCredentialResponseJson response;
		if (theRequest.getPartitionId() != 0) {
			response = determinePartitionById(theRequest.getPartitionId());
		} else if (!StringUtils.isBlank(theRequest.getPartitionName())){
			response = determinePartitionByName(theRequest.getPartitionName());
		} else {
			throw new InvalidRequestException("No way to identify this partition!");
		}
		/*
		 * If any additional SQL should be executed in the database when it is
		 * first initialized, it may be added here. The example below shows a
		 * no-op query which does not do anything (and this should be removed
		 * in a real deployment), but you could replace this with a script that
		 * sets permissions or adds placeholders.
		 */
		response.addDatabaseInitializationStatement("SELECT 1");
		return response;
	}

	/**
	 * You can elect to determine data source selection by tenant name, instead of by partition ID, if you so desire.
	 */
	private MegaScaleCredentialResponseJson determinePartitionByName(String partitionName) {
		/*
		 * In this example, we'll use one connection URL for tenants "TENANT-A" & "TENANT-C", and
		 * a different connection URL for partitions "TENANT-B" & "TENANT-D". You can group partitions
		 * any way you would like.
		 */
		MegaScaleCredentialResponseJson retVal = new MegaScaleCredentialResponseJson();
		switch (partitionName) {
			case "TENANT-A":
			case "TENANT-C":
				setConnectionToFirstDatabase(retVal);
				break;
			case "TENANT-B":
			case "TENANT-D":
				setConnectionToSecondDatabase(retVal);
				break;
			default:
				// Shouldn't happen, and will cause an error if it does
		}
		return retVal;

	}

	private void setConnectionToSecondDatabase(MegaScaleCredentialResponseJson retVal) {
		retVal.setDatabaseUrl("jdbc:h2:mem:ds2");
		retVal.setDatabaseUsername("SA");
		retVal.setDatabasePassword("SA");
	}

	private void setConnectionToFirstDatabase(MegaScaleCredentialResponseJson retVal) {
		retVal.setDatabaseUrl("jdbc:h2:mem:ds1");
		retVal.setDatabaseUsername("SA");
		retVal.setDatabasePassword("SA");
	}


	private MegaScaleCredentialResponseJson determinePartitionById(int partitionId) {
		/*
		 * In this example, we'll use one connection URL for partitions 1 & 3, and
		 * a different connection URL for partitions 2 & 4. You can group partitions
		 * any way you would like.
		 */
		MegaScaleCredentialResponseJson retVal = new MegaScaleCredentialResponseJson();
		switch (partitionId) {
			case 1:
			case 3:
				setConnectionToFirstDatabase(retVal);
				break;
			case 2:
			case 4:
				setConnectionToSecondDatabase(retVal);
				break;
			default:
				// Shouldn't happen, and will cause an error if it does
		}
		/*
		 * If any additional SQL should be executed in the database when it is
		 * first initialized, it may be added here. The example below shows a
		 * no-op query which does not do anything (and this should be removed
		 * in a real deployment), but you could replace this with a script that
		 * sets permissions or adds placeholders.
		 */
		retVal.addDatabaseInitializationStatement("SELECT 1");

		return retVal;
	}

}
