/*-
 * #%L
 * Smile CDR - CDR
 * %%
 * Copyright (C) 2016 - 2024 Smile CDR, Inc.
 * %%
 * All rights reserved.
 * #L%
 */
package com.smilecdr.demo.server.r4;

import ca.uhn.fhir.interceptor.api.Hook;
import ca.uhn.fhir.interceptor.api.Interceptor;
import ca.uhn.fhir.interceptor.api.Pointcut;
import ca.uhn.fhir.jpa.api.dao.DaoRegistry;
import ca.uhn.fhir.jpa.api.dao.IFhirResourceDao;
import ca.uhn.fhir.jpa.api.model.DaoMethodOutcome;
import ca.uhn.fhir.mdm.api.MdmConstants;
import ca.uhn.fhir.mdm.api.MdmMatchResultEnum;
import ca.uhn.fhir.mdm.model.mdmevents.MdmLinkEvent;
import ca.uhn.fhir.mdm.model.mdmevents.MdmLinkJson;
import ca.uhn.fhir.rest.client.api.IGenericClient;
import ca.uhn.fhir.rest.server.TransactionLogMessages;
import ca.uhn.fhir.rest.server.exceptions.BaseServerResponseException;
import ca.uhn.fhir.rest.server.messaging.BaseResourceMessage;
import ca.uhn.fhir.rest.server.messaging.ResourceOperationMessage;
import ca.uhn.fhir.util.StopWatch;
import org.hl7.fhir.instance.model.api.IBaseResource;
import org.hl7.fhir.r4.model.CodeableConcept;
import org.hl7.fhir.r4.model.IdType;
import org.hl7.fhir.r4.model.Identifier;
import org.hl7.fhir.r4.model.Patient;
import org.hl7.fhir.r4.model.Person;
import org.hl7.fhir.r4.model.Reference;
import org.hl7.fhir.r4.model.ResourceType;
import org.hl7.fhir.r4.model.Task;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import jakarta.annotation.PostConstruct;
import java.nio.charset.StandardCharsets;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Collectors;

/**
 * This sample interceptor is intended to be used with MDM module.
 * MDM module configuration 'Interceptor Bean Types' should be set to 'com.smilecdr.demo.server.r4.TestServerAppCtx'
 * to test it in Smile CDR, so appropriate dependencies could be passed to its constructor.
 * <p>
 * The sample interceptor has two goals, when operating on Patient resources.
 * </p><p>
 * The first goal is to make sure Patient resources have US SSN identifier,
 * which is treated as an important value in our sample scenario,
 * before the MDM module searches for candidate Patient resources and tries to create links.
 * If the operated Patient resource is missing a US SSN identifier,
 * the interceptor will then create a Task resource in the internal FHIR repository to facilitate tracking
 * these Patient resources that need to be updated.
 * </p><p>
 * The second goal is to create a Person resource on an external FHIR repository,
 * based on the golden Patient resource, for newly created Patient resources,
 * once the MDM module linked the created Patient resources to a golden Patient resource.
 * </p>
 */
@Interceptor
public class PatientMDMInterceptor {

	// let's add some entries in smile.log!
	private static final Logger ourLog = LoggerFactory.getLogger(PatientMDMInterceptor.class);

	private final static String IDENTIFIER_US_SSN_SYSTEM = "http://hl7.org/fhir/sid/us-ssn";

	// used to access internal / local FHIR repository
	private final IGenericClient myInternalFhirClient;

	// used to access a potential external / remote FHIR repository
	private final IGenericClient myExternalFhirClient;

	// used to access local FHIR repository directly with Dao, see important note about it in TestServerAppCtx
	private final DaoRegistry myDaoRegistry;

	private IFhirResourceDao<Patient> myPatientDao;

	private IFhirResourceDao<Task> myTaskDao;

	public PatientMDMInterceptor(
		IGenericClient theInternalFhirClient,
		IGenericClient theExternalFhirClient,
		DaoRegistry theDaoRegistry) {

		myInternalFhirClient = theInternalFhirClient;
		myExternalFhirClient = theExternalFhirClient;
		myDaoRegistry = theDaoRegistry;
	}

	@PostConstruct
	public void init() {
		// we don't need to retrieve the DAO from the registry at each usage, so do it once in a PostConstruct method
		ourLog.info("Initialize Patient MDM Interceptor");
		myPatientDao = myDaoRegistry.getResourceDao(Patient.class);
		myTaskDao = myDaoRegistry.getResourceDao(Task.class);
	}

	/**
	 * Interception method, called before MDM module searches for candidate resources and tries to create links.
	 * @param theResource Operated resource, which MDM will analyze after interception
	 */
	@Hook(value = Pointcut.MDM_BEFORE_PERSISTED_RESOURCE_CHECKED)
	public void mdmBeforePersistedResourceChecked(IBaseResource theResource) {

		if (!ResourceType.Patient.toString().equals(theResource.fhirType())) {
			// safeguard, quickly exit if we intercepted MDM call on something other than Patient resource
			// which shouldn't happen if MDM is properly configured
			return;
		}

		ourLog.info("Interceptor MDM_BEFORE_PERSISTED_RESOURCE_CHECKED - started");
		StopWatch stopWatch = new StopWatch();
		try {
			Patient patient = (Patient) theResource;
			Optional<Identifier> ssnIdentifier = patient.getIdentifier().stream()
				.filter(identifier -> IDENTIFIER_US_SSN_SYSTEM.equals(identifier.getSystem()))
				.findFirst();
			if (ssnIdentifier.isEmpty()) {
				Task task = buildTaskForPatientWithoutSsn(patient);

				ourLog.info("Saving\n{}\n",
					myInternalFhirClient.getFhirContext().newJsonParser().setPrettyPrint(true)
						.encodeResourceToString(task));

				DaoMethodOutcome daoMethodOutcome = myTaskDao.create(task);
				// we could have used the internal FhirClient here instead of calling the Dao
				// if we had wanted more standard/robust code
				// myInternalFhirClient
				//	.create()
				//	.resource(task)
				//	.execute();
				if (daoMethodOutcome.getCreated()) {
					ourLog.info("Task successfully created\n{}\n",
						myInternalFhirClient.getFhirContext().newJsonParser().setPrettyPrint(true)
							.encodeResourceToString(daoMethodOutcome.getResource()));
				}
			}
		}
		finally {
			ourLog.info("Interceptor MDM_BEFORE_PERSISTED_RESOURCE_CHECKED - ended, execution took {}", stopWatch);
		}
	}

	/**
	 * Interception method, called after MDM module created links for the operated resource
	 * according to the MDM rule definition provided in MDM module.
	 * @param theResourceOperationMessage The information about the operated resource
	 * @param theTransactionLogMessages The Transaction log messages about the operated resource
	 * @param theMdmLinkEvent The MDM links information, after MDM module analyzed the operated resource
	 */
	@Hook(value = Pointcut.MDM_AFTER_PERSISTED_RESOURCE_CHECKED)
	public void mdmAfterPersistedResourceChecked(
		ResourceOperationMessage theResourceOperationMessage,
		TransactionLogMessages theTransactionLogMessages,
		MdmLinkEvent theMdmLinkEvent) {

		IBaseResource resource = theResourceOperationMessage.getPayload(myInternalFhirClient.getFhirContext());
		if (!ResourceType.Patient.toString().equals(resource.fhirType())) {
			// safeguard, quickly exit if we intercepted MDM call on something other than Patient resource
			// which shouldn't happen if MDM is properly configured
			return;
		}

		ourLog.info("Interceptor MDM_AFTER_PERSISTED_RESOURCE_CHECKED - started");
		StopWatch stopWatch = new StopWatch();
		try {
			// verify the operation done on the operated resource,
			// as in this demo, we want to check MDM links for newly created Patient resource only
			if (BaseResourceMessage.OperationTypeEnum.CREATE.equals(theResourceOperationMessage.getOperationType())) {

				for (MdmLinkJson mdmLinkJson : theMdmLinkEvent.getMdmLinks()) {
					String goldenResourceId = mdmLinkJson.getGoldenResourceId();

					if (MdmMatchResultEnum.MATCH.equals(mdmLinkJson.getMatchResult())) {
						// it means the newly created Patient resource is linked to a golden resource
						// as a matching resource, according to the MDM rule definition provided in MDM module
						Patient goldenResource = getGoldenResource(goldenResourceId);
						if (goldenResource != null) {
							// build a unique Person resource based on golden resource information
							Person person = buildPersonFromGoldenResource(goldenResource);

							ourLog.info("Sending to external FHIR repository\n{}\n",
								myExternalFhirClient.getFhirContext().newJsonParser().setPrettyPrint(true)
									.encodeResourceToString(person));

							// and send that Person resource to our external FHIR repository
							// using PUT {external-base-url}/Person/{person-fhir-id}
							// so this will either update or create the Person resource if not present
							myExternalFhirClient
								.update()
								.resource(person)
								.withId(person.getIdElement().getIdPart())
								.execute();
						}
					}
					else if (MdmMatchResultEnum.POSSIBLE_MATCH.equals(mdmLinkJson.getMatchResult())) {
						// in this case, we could create a task for someone to settle the possible match
						// using $mdm-update-link operation!
						ourLog.info("MDM Possible match link for golden resource {}", goldenResourceId);
					}
				}
			}
		}
		finally {
			ourLog.info("Interceptor MDM_AFTER_PERSISTED_RESOURCE_CHECKED - ended, execution took {}", stopWatch);
		}
	}

	private Task buildTaskForPatientWithoutSsn(Patient thePatient) {

		Reference patientReference = new Reference(thePatient);

		return new Task()
			.setStatus(Task.TaskStatus.DRAFT)
			.setIntent(Task.TaskIntent.PLAN)
			.setCode(new CodeableConcept().setText("Missing required information"))
			.setFocus(patientReference)
			.setFor(patientReference)
			.setDescription("Patient without Social Security Number; patient should be contacted to get that important information");
	}

	private Patient getGoldenResource(String theGoldenResourceId) {

		Patient goldenResource = null;
		try {
			// golden resources are like any other Patient resources
			// but with meta.tag entries and an additional identifier
			// (having 'http://hapifhir.io/fhir/NamingSystem/mdm-golden-resource-enterprise-id' system)
			// MDM module keeps the golden record updated according to MDM Survivorship Script
			// (usually, with the latest Patient information, but it is configurable)
			goldenResource = myPatientDao.read(new IdType(theGoldenResourceId));

			// we could have used the internal FhirClient here instead of calling the Dao
			// if we had wanted more standard/robust code
			// goldenResource = myInternalFhirClient
			//	.read()
			//	.resource(Patient.class)
			//	.withId(goldenResourceId)
			//	.execute();

		} catch (BaseServerResponseException theBaseServerResponseException) {
			ourLog.error("Unable to get golden resource " + theGoldenResourceId,
				theBaseServerResponseException);
		}
		return goldenResource;
	}

	private Person buildPersonFromGoldenResource(
		Patient theGoldenResource) {

		// build Person resource from Patient golden record
		// here, for demonstration purposes, we don't want to forward any internal FHIR id
		// to the external FHIR repository
		// so create the same Person FHIR id for the same Patient golden record FHIR id
		// to make sure we don't create multiple Person resources for the same Patient golden record
		// as it is a purposed requirement on the external FHIR repository
		UUID personFhirIdUUID = UUID.nameUUIDFromBytes(
			theGoldenResource.getIdElement().getIdPart().getBytes(StandardCharsets.UTF_8)
		);

		Person person = new Person();
		person.setId(personFhirIdUUID.toString());
		// filter out the additional internal identifier
		// (having 'http://hapifhir.io/fhir/NamingSystem/mdm-golden-resource-enterprise-id' system)
		// from golden record
		person.setIdentifier(theGoldenResource.getIdentifier().stream()
			.filter(identifier -> MdmConstants.HAPI_ENTERPRISE_IDENTIFIER_SYSTEM.equals(identifier.getSystem()))
			.collect(Collectors.toList()));
		if (theGoldenResource.hasActive()) {
			person.setActive(theGoldenResource.getActive());
		}
		person.setBirthDate(theGoldenResource.getBirthDate());
		person.setGender(theGoldenResource.getGender());
		person.setName(theGoldenResource.getName());
		person.setAddress(theGoldenResource.getAddress());
		person.addLink().setTarget(
			new Reference()
				.setType(ResourceType.Patient.toString())
				// dao_config.allow_external_references.enabled must be enabled
				// 'Allow External References Enabled' in 'Persistence' module
				.setReference(myExternalFhirClient.getServerBase()
					+ (!myExternalFhirClient.getServerBase().endsWith("/") ? "/" : "")
					+ theGoldenResource.fhirType() + "/" + theGoldenResource.getIdPart())
				.setDisplay("Golden resource reference in source FHIR repository")
			)
			// According to the MDM rule definition, this is the same person
			.setAssurance(Person.IdentityAssuranceLevel.LEVEL3);
		return person;
	}
}
