/*-
 * #%L
 * Smile CDR - CDR
 * %%
 * Copyright (C) 2016 - 2024 Smile CDR, Inc.
 * %%
 * All rights reserved.
 * #L%
 */
package com.smilecdr.demo.server.r4;

import ca.uhn.fhir.context.FhirContext;
import ca.uhn.fhir.jpa.api.dao.DaoRegistry;
import ca.uhn.fhir.rest.client.api.IGenericClient;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;

/**
 * This sample interceptor configuration is intended to be used with MDM module.
 * MDM module configuration 'Interceptor Bean Types' should be set to 'com.smilecdr.demo.server.r4.TestServerAppCtx'
 * to test PatientMDMInterceptor in Smile CDR.
 * <p>
 * Sample MDM module configuration
 * </p><p>
 * A sample 'MDM Rule Definition Script' is provided in resources/patient.mdm.interceptor/sample.mdm.rule.definition.json
 * and 'MDM Survivorship Script' is provided in resources/patient.mdm.interceptor/sample.mdm.survivorship.script.js
 * that could be used to test PatientMDMInterceptor.
 * See <a href="https://smilecdr.com/docs/mdm/quickstart.html">MDM Quickstart</a>
 * and <a href="https://smilecdr.com/docs/mdm/mdm_survivorship.html">MDM Survivorship rules</a>
 * for more information about MDM configuration.
 * </p>
 */
@Configuration
public class TestServerAppCtx {

	// base url to our local/internal FHIR repository
	private static final String INTERNAL_FHIR_BASE_URL = "http://localhost:8000/";

	// base url to an external FHIR repository
	// it could be another Smile CDR FHIR Endpoint hosted in the cloud, for example,
	// or a second FHIR endpoint configured in the same Smile CDR instance.
	// For the purposes of this demo, and to make it simpler, we will use the same endpoint as our internal one
	// and name it external in the code.
	// The important part is that we will create two distinct FHIR clients.
	private static final String EXTERNAL_FHIR_BASE_URL = "http://localhost:8000/";

	/**
	 * Important note about DaoRegistry
	 * <p>
	 * It is not the best practice to use Dao to access data, as they are highly tied to Hapi-FHIR / Smile CDR version.
	 * The interfaces and/or behaviour can change quite a lot between versions,
	 * so there should be no expectation on compatibility between versions used.
	 * It is much safer to use a FHIR client for long term use to retrieve/send data from/to a FHIR repository,
	 * (IGenericClient created from FhirContext.newRestfulGenericClient())
	 * as the interfaces and/or behaviour are more guaranteed by the FHIR specification.
	 * But in some specific cases, a Dao can be used. This example demonstrates how to use the Dao.
	 * Also, using Dao could allow users to access data that might not be permitted normally,
	 * depending on the user permissions. So it needs to be planned carefully.
	 * </p>
	 * Implementation note
	 * <p>
	 * Smile CDR will provide an instance of DaoRegistry via dependency injection,
	 * so it is not required to manually configure it / add it in Spring Context
	 * with @Bean annotation.
	 * </p>
	 */
	@Autowired
	private DaoRegistry myDaoRegistry;

	@Primary
	@Bean(name = "interceptorFhirContext")
	public FhirContext getFhirContext() {
		return FhirContext.forR4();
	}

	@Bean(name = "internalFhirClient")
	public IGenericClient getInternalFhirClient(
		@Qualifier("interceptorFhirContext") FhirContext theFhirContext) {
		return theFhirContext.newRestfulGenericClient(INTERNAL_FHIR_BASE_URL);
	}

	@Bean(name = "externalFhirClient")
	public IGenericClient getExternalFhirClient(
		@Qualifier("interceptorFhirContext") FhirContext theFhirContext) {
		return theFhirContext.newRestfulGenericClient(EXTERNAL_FHIR_BASE_URL);
	}

	@Bean(name = "patientMDMInterceptor")
	public PatientMDMInterceptor getPatientMDMInterceptor(
		@Qualifier("internalFhirClient") IGenericClient theInternalFhirClient,
		@Qualifier("externalFhirClient") IGenericClient theExternalFhirClient) {

		return new PatientMDMInterceptor(theInternalFhirClient, theExternalFhirClient, myDaoRegistry);
	}
}
